/*
 Find symplectic fillings of links of quotient surface singularities
 */
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <sstream>
using namespace std;

ifstream in("SingularitiesList.txt"); // Open for reading
// Each line in the input file "SingularitiesList.txt should have the format
// "Type D.D C_1.C_1 ... C_k.C_k", where "1" denotes Type (3,2) and "2" denotes 
// Type (3,1).
ofstream out("FillingsList.txt"); // Open for writing

class Config // Internal representation of a compactifying divisor
{
public:
    int D; // D squared
    int k; // Length of the string C_1,...,C_k
    vector<int> C; // C[i] = C_{i+1}.C_{i+1}
    int i; // If i != -1, there exists a (-1)-curve intersecting B and C_{i+1}
    int j; // If j != -1, there exists a (-1)-curve intersecting D and C_{j+1}
    Config(int DD, int kk, int ii, int jj); // Constructor
    Config(int DD, int kk); // Constructor
    void print(); // Print Config without parentheses
    void pprint();  // Print Config with parentheses
};

Config::Config(int DD, int kk, int ii, int jj)
{
    D = DD;
    k = kk;
    i = ii;
    j = jj;
    for(int s = 0; s < k; s++) C.push_back(0);
}

Config::Config(int DD, int kk)
{
    D = DD;
    k = kk;
    i = -1;
    j = -1;
    for(int s = 0; s < k; s++) C.push_back(0);
}

void Config::print()
{
    out << D << ",";
    for(int s = 0; s < (k-1); s++) out << C[s] << ",";
    if(k > 0) out << C[k-1];
}

void Config::pprint()
{
    out << "(" << D << ",";
    for(int s = 0; s < (k-1); s++) out << C[s] << ",";
    if(k > 0) out << C[k-1];
    out << ")" << "\n";
}

class Exc // Class to represent the exceptional curves intersecting C_1,...,C_k
{
public:
    int k;
    vector<int> E; // E[i] denotes the number of (-1)-curves intersecting C_{i+1}
    Exc(int kk); // Constructor
    void print();
};

Exc::Exc(int kk)
{
    k=kk;
    for(int s = 0; s < k; s++) E.push_back(0);
}

void Exc::print()
{
    int ctr = -1;
    for(int i = 0; i < k; i++) {
        if(E[i] > 0) {
            ctr++;
            if(ctr > 0) out << ",";
        }
        if(E[i] == 1) out << i + 1;
        if(E[i] > 1) out << E[i] << "x" << i + 1;
    }
}

Config GetConfig(string s) // Construct a Config from a string which encodes
// a compactifying divisor (the format should be as in the input file).
{
    int DD = 0;
    int nn;
    vector<int> CC;
    stringstream stream(s);
    stream >> nn;  // Discard first entry, which is Case
    stream >> DD;
    while(stream >> nn) CC.push_back(nn);
    Config X(DD, CC.size());
    for(int s = 0; s < X.k; s++)
        X.C[s] = CC[s]; 
    return X;
}

Config BlowDown(Config X, Exc& S) // Blow down all (-1)-curves in S intersecting
// compactifying divisor and (-1)-curves corresponding to i and j, if they exist.
{
    for(int s = 0; s < X.k; s++)
        X.C[s]=(X.C[s] + S.E[s]); // Blow down (-1)-curves in S
    if(X.i != -1) {
        X.C[X.i]++; X.C[X.j]++;
        X.D++; // Blow down (-1)-curves corresponding to i and j, if they exist
    }
    return X;
}

Config Collapse(Config& X, int r) // Blow down (-1)-curve C_{r + 1}.
// Note that we assume that r != i, and that r != j unless j = k - 1.
{
    Config Y(X.D, X.k - 1);
    if(r == 0) {
        Y.D++;
        if(Y.k > 0) {
            Y.C[0] = X.C[1] + 1;
            for(int s = 1; s < Y.k; s++) Y.C[s] = X.C[s + 1];
        }
        if(X.i > -1) {
            Y.i = X.i - 1;
            Y.j = X.j - 1;
        }
    }
    if(r > 0 && r < X.k - 1) {
        for(int s = 0; s < r - 1; s++) Y.C[s] = X.C[s];
        Y.C[r - 1] = X.C[r - 1] + 1;
        Y.C[r] = X.C[r + 1] + 1;
        for(int s = r + 1; s < Y.k; s++) Y.C[s] = X.C[s + 1];
        if(r < X.i) {
            Y.i = X.i - 1;
            Y.j = X.j - 1;
        }
        if(r > X.i && r < X.j) { // r cannot equal X.i
            Y.i = X.i;
            Y.j = X.j - 1;
        }
        if(r > X.j) { // r cannot equal X.j since r < X.k - 1
            Y.i = X.i;
            Y.j = X.j;
        }
    }
    if(r == X.k - 1 && r != 0) {
        for(int s = 0; s < Y.k - 1; s++) Y.C[s] = X.C[s];
        Y.C[Y.k - 1] = X.C[Y.k - 1] + 1;
        Y.i = X.i;
        if(X.j == X.k - 1) {
            Y.D++;
            Y.j = X.j - 1;
        }
        else Y.j = X.j;
    }
    return Y;
}

int FindExceptional(Config& X) // Returns s if C_{s + 1} is a (-1)-curve which 
// can be blown down preserving the class of configurations; returns -1 if no 
// such curve exists.
{
    for(int s = 0; s < X.k; s++)
        if(X.C[s] == -1 && s!= X.i && (s != X.j || s == X.k - 1)) return s;
    return -1;
}

int AdmissibilityTest(int Case, Config& X) // Tests whether a compactifying 
// divisor of a given type is admissible, i.e., whether it can be blown down 
// to a standard configuration. Returns 0 if the compactifying divisor is not 
// admissible, and 1 (resp. 2) if the compactifying divisor is admissible 
// (resp. and the sing. is of Type (3,2), and D.D becomes 8 after blowing down).
{
    switch (Case) {
        case 0 : if(X.k == 0) {
                if(X.D == 8) return 2;
                else if(X.D < 10) return 1;
                else return 0;
            }
            break;
        case 1 : if(X.k == 0) {
                if(X.D < 9) return 1;
                else return 0;
            }
            break;
        case 2 : if(X.k == 1) {
                if(X.D < 9 && X.C[0] == 0) return 1;
                else return 0;
            }
            break;
    }
    int r = FindExceptional(X);
    if(r == -1 && X.k > 0) return 0; // If no suitable curve in X can be blown 
                                     // down
    Config Y = Collapse(X, r);
    return AdmissibilityTest(Case, Y); // If a suitable curve can be blown down
}

int Increment(Config& X, Exc& S, int& ctr) // Finds the next possible collection 
// of exceptional divisors intersecting C_1,...,C_k
{
    if(S.E[ctr] > -X.C[ctr] - 2 && ctr == X.k - 1) return 0;
    if(S.E[ctr] > -X.C[ctr] - 2 && ctr < X.k - 1) {
        S.E[ctr] = 0;
        ctr++;
        return Increment(X, S, ctr);
    }
    else {
        S.E[ctr]++;
        ctr = 0;
        return 1;
    }
}

void EnumerateFillings(int Case, int& num, Config& X) // List fillings of a  
// given singularity of a given type: Case 0 = Type (3,2) singularity; Case 1 = 
// Type (3,1) singularity Case I; Case 2 = Type (3,1) singularity Case II. 
// Fillings are enumerated by the variable num.
{   
    Exc S(X.k);
    int ctr = 0;
    int num_0 = num;
    switch (Case) {
        case 0 : do 
            {
                Config Y = BlowDown(X,S);
                int N = AdmissibilityTest(Case, Y);
                if(N == 1) {
                    out << num << ". (";
                    num++;
                    X.print();
                    out << ";";
                    S.print();
                    out << "), CP^2" << endl;
                }
                if(N == 2) {
                    out << num << ". (";
                    num++;
                    X.print();
                    out << ";";
                    S.print();
                    out << "), CP^2" << endl;
                    out << num << ". (";
                    num++;
                    X.print();
                    out << ";";
                    S.print();
                    out << "), CP^1 x CP^1" << endl;
                }
            } while(Increment(X, S, ctr));
            break;
        case 1 : do 
            {
                Config Y = BlowDown(X,S);
                int N = AdmissibilityTest(Case, Y);
                if(N == 1) {
                    out << num << ". (";
                    num++;
                    X.print();
                    out << ";";
                    S.print();
                    out << ")" << endl;
                }
            } while(Increment(X, S, ctr));
            break;
        case 2 : for(X.i = 0; X.i < X.k; ++X.i)
                for(X.j = X.i; X.j < X.k; ++X.j) {
                    Exc S(X.k);
                    ctr = 0;
                    do 
                    {
                        Config Y = BlowDown(X,S);
                        if(AdmissibilityTest(Case, Y) == 1) {
                            out << num << ". (";
                            num++;
                            X.print();
                            out << ";" << X.i + 1 << "," << X.j + 1 << ";";
                            S.print();
                            out << ")" << endl;
                        }
                    } while(Increment(X, S, ctr));
                }
            if(num == num_0) out << "None." << endl;
            break;
    }
}

int main()
{
    string s;
    int num = 1;
    while(getline(in, s)) {
        int Case = 0;
        stringstream stream(s);
        stream >> Case;
        Case--;
        Config Z = GetConfig(s);
        if(Case != 0 && Case != 1) {
            out << "Type not recognized." << endl;
        }
        if(Z.D < 1) {
            out << "Unknown or missing data." << endl;
            Case = -1; // Don't try to compute fillings
        }
        if(Z.D > 9) {
            out << "Impossible." << endl;
            Case = -1;
        }
        if(Case == 0) {
            out << "Type (3,2) singularity ";
            Z.pprint();
        }
        if(Case == 1) {
            out << "Type (3,1) singularity ";
            Z.pprint();
        }
        if(Z.k == 0 && (Case == 0 || Case == 1) && Z.D < 10 && Z.D > 0) {
            out << num << ". " << "There is a unique filling." << endl;
            num++;
            Case = -1; 
        }
        switch (Case) {
            case 0 : out << "The fillings are:" << endl;
                EnumerateFillings(Case, num, Z);
                break;
            case 1 : out << "Case I fillings:" << endl;
                EnumerateFillings(Case, num, Z);
                Case++;
                out << "Case II fillings:" << endl;
                EnumerateFillings(Case, num, Z);
                break;
        }
        out << endl;
    }
}
